from django.db import models
from django.db.utils import ProgrammingError
from django.forms import model_to_dict


class SingletonModel(models.Model):
    """Singleton Django Model"""

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        self.__class__.objects.exclude(id=self.id).delete()
        super(SingletonModel, self).save(*args, **kwargs)
        return self

    @classmethod
    def get(cls, attr: str = None):
        try:
            obj = cls.objects.get()
        except cls.DoesNotExist:
            obj = cls.objects.create()
        except ProgrammingError as e:
            return
        if attr:
            return getattr(obj, attr)
        return obj

    @classmethod
    def update(cls, **kwargs):
        obj = cls.get()
        for k, v in kwargs.items():
            setattr(obj, k, v)
        obj.save()
        return obj

    @classmethod
    def to_dict(cls):
        instance = cls.get()
        return model_to_dict(instance)



class ConditionalQuerysetMixin:
    """
    ConditionalQuerysetMixin gives you an ability to have multiple `get_queryset`
    methods for every needs.

    Querysets mapping
    `querysets` - is a dictionary of named querysets. You can provide different
    querysets for any action.

    Named queryset method
    Use named methods like `get_name_queryset`, where `name` is an any name.

    ViewSet actions
    If ConditionalQuerysetMixin can't find a mapped or named queryset, it will try to
    find a method with a ViewSet action instead of name.
    E.g: `get_list_queryset`, `get_update_queryset`, get_custom_action_queryset.
    """
    querysets = {}

    def get_queryset(self, **kwargs):
        queryset = super().get_queryset()
        name = kwargs.get('name', self.action)
        if name:
            queryset = self.querysets.get(name, queryset)
        method = self.try_method(f"get_{name}_queryset")
        if method:
            queryset = method(queryset)
        return queryset

    def try_method(self, method_name: str):
        return getattr(self, method_name, None)



from django.conf import settings


def get_settings():
    default = dict(
        run_gears_server_up_tasks=[],
        run_gears_server_down_tasks=[],
    )

    conf = getattr(settings, "GEARS", {})

    return {
        attr_name: conf.get(attr_name, default_value)
        for attr_name, default_value in default.items()
    }



from rest_framework_simplejwt.views import TokenObtainPairView

from ..serializers.jwt import JWTObtainPairSerializer


class JWTObtainPairView(TokenObtainPairView):
    """
    This is a helper view for handling the JWT gears.
    Use it, if you need a power of JWT tokens with a payload inside.
    """

    serializer_class = JWTObtainPairSerializer



from ..exceptions.views import GearsViewException


class ServiceDataViewSetMixin(object):
    def get_renderer_context(self):
        context = super().get_renderer_context()
        try:
            context.update({
                'service': self.get_service_data(self.request)
            })
        except AttributeError:
            GearsViewException("The method get_service_data not found.")
        return context



from rest_framework.response import Response


class SummaryPaginationMixin:
    """
    This mixin for a DRF pagination class provides a mechanism to add some additional
    summary data. Use it with some standard pagination class like PageNumberPagination
    or LimitOffsetPagination. You should add the method named `get_pagination_summary`
    for ViewSet which returns a dictionary with needed information.
    """

    def get_summary(self, view, queryset, request):
        if view and hasattr(view, 'get_pagination_summary'):
            return view.get_pagination_summary(request, queryset)

    def paginate_queryset(self, queryset, request, view=None):
        self.summary = self.get_summary(view, queryset, request)
        return super().paginate_queryset(queryset=queryset, request=request, view=view)

    def get_paginated_response(self, data):
        response = super().get_paginated_response(data)
        data = response.data
        data['summary'] = self.summary
        return Response(data)



class PermissionsMixin:
    """
    PermissionsMixin ...
    """
    permission_classes = []
    permission_classes_as_global = True  # use permission_classes globally if True
    default_name = 'default'
    permissions = {}

    def get_permission_classes(self):
        global_classes = []
        if self.permission_classes_as_global:
            global_classes.extend(self.permission_classes)

        classes = self.permissions.get(self.action) \
                        or self.permissions.get(self.default_name)
        if classes:
            global_classes.extend(classes)

        return global_classes

    def get_permissions(self):
        return [permission() for permission in self.get_permission_classes()]



from typing import Union

from rest_framework.renderers import JSONRenderer

from .types import Response


class APIRenderer(JSONRenderer):
    pagination_result_field: str = 'results'
    errors_field: str = 'errors'

    def process(self, data, renderer_context):
        # request = renderer_context['request']
        response = renderer_context['response']
        success = not response.exception
        pagination = self.get_pagination(data)
        return Response(
            success=success,
            status_code=response.status_code,
            pagination=pagination,
            errors=self.get_errors(success, data, renderer_context),
            data=self.get_data(success, pagination, data, renderer_context),
            service_data=self.get_service_data(
                success, pagination, data, renderer_context,
            ),
        )

    def render(self, data, accepted_media_type=None, renderer_context=None):
        data = self.process(data, renderer_context)
        data = data.__dict__
        ret = super().render(data, accepted_media_type, renderer_context)
        return ret

    def get_errors(self, success: bool, data, renderer_context) -> list:
        if success:
            return []
        return data[self.errors_field]

    def get_pagination(self, data) -> Union[dict, None]:
        try:
            pagination_data = {
                k: v for k, v in data.items() if k != self.pagination_result_field
            }
        except AttributeError:
            pagination_data = {}
        if pagination_data.get('count') is not None:
            return pagination_data

    def get_data(
            self, success: bool, pagination, data, renderer_context,
    ) -> Union[dict, list, None]:
        if not success:
            return
        if not pagination:
            return data
        return data[self.pagination_result_field]

    def get_service_data(
            self, success: bool, pagination, data, renderer_context,
    ) -> Union[dict, list, None]:
        if not success:
            return
        return renderer_context.get('service')



from typing import List, Tuple, Set

from django.db import models


class OnChangeModel(models.Model):
    """
    Handle fields edition logic by the custom methods.
    You have two options which could works together. The first one is a list of model
    fields, which will be processed by the main on_change method. By default, it does
    nothing, so take care to write your own on_change method.
    The second one is a multiple custom methods, a single for any field you want
    to handle up. Name these methods like `on_change_<field_name>`. BTW, you're able to
    rename this prefix by the `on_change_prefix`.
    Keep on mind, if you use the on_change_field method, it will be executed first, and
    the main on_change method will be executed last.
    You can manage the main on_change method from the on_change_field methods by
    returning the status value. It shows your desire the main on_change method working
    on this field or not.

    on_change_fields -- A list of field for processing by OnChangeModel. All of them
    will be processed by the on_change method.

    Create a method with name like 'on_change_<field_name>' for handling fields with
    separated method.

    on_change_prefix -- a prefix for methods naming.
    origin_prefix -- a prefix for attributes where serves the origin values.
    """

    # It would be nice to make a decorator @on_change(field1, field2...)

    on_change_fields = ()
    on_change_prefix = 'on_change_'
    post_change_fields = ()
    post_change_prefix = 'post_change_'
    origin_prefix = '__origin_'

    class Meta:
        abstract = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._on_change_fields = self.get_on_change_fields()
        self._post_change_fields = self.get_post_change_fields()
        fields = self._on_change_fields + self._post_change_fields
        for name in set(fields):
            setattr(self, self.get_origin_name(name), getattr(self, name, None))
        setattr(self, self.get_origin_name('counter'), 0)

    def save(self, *args, **kwargs):
        self.process_changed_fields(self.on_change_prefix, *args, **kwargs)
        super().save(*args, **kwargs)
        self._increase_counter()
        self.process_changed_fields(self.post_change_prefix, *args, **kwargs)
        self.after_save()

    def execute_change_method(
            self, prefix, origin_name, origin_value, name, value,
    ) -> bool:
        """
        This method executes the custom model's on_change method.
        On_change methods could impact on the values will be set. So take care
        to return the status of its work.
        The status is a boolean indicator of the on_change method, where True means
        this field should be passed to the main on_change method, and False
        means it shouldn't.
        :param prefix: maybe on_change or post_change
        :param origin_name: the prefixed name of an attribute where the original value
        is stored
        :param origin_value: the original value of the field
        :param name: a name of the field
        :param value: new value of the field
        :return: status - must be processed by the main on_change method or must not.
        """
        method = self._get_method(prefix, name)
        if not method:
            return True  # must be processed by the main on_change method
        _status = method(origin_value, value)
        # This is a compatibility part. Previous lib version might return nothing.
        _status = True if _status is None else _status
        # set new value as original value preventing extra method execution
        # setattr(self, origin_name, value)
        return _status

    def get_origin_name(self, name):
        return f'{self.origin_prefix}{name}'

    def get_origin_value(self, origin_name):
        return getattr(self, origin_name)

    def process_changed_fields(self, prefix, *args, **kwargs):
        update_fields = set(kwargs.get('update_fields', []))
        changed_fields = set(self.get_changed_fields(prefix))
        change_attrs = []
        for origin_name, origin_value, name, value in changed_fields:
            if update_fields and name not in update_fields:
                # exclude all another fields if we use the `update_fields` option
                continue
            attrs = origin_name, origin_value, name, value
            status = self.execute_change_method(prefix, *attrs)
            if status is True:
                change_attrs.append(attrs)
        self.batch_change(prefix, change_attrs)

    def after_save(self):
        """Customize it on your taste"""
        pass

    def batch_change(self, prefix: str, fields: List[Tuple]):
        if prefix == self.on_change_prefix:
            return self.on_change(fields)
        elif prefix == self.post_change_prefix:
            return self.post_change(fields)

    def on_change(self, fields: List[Tuple]):
        """
        This method allows to implement the logic based on the whole
        changed fields list. It works only with the field defined in the
        `on_change_fields` attribute. If you need to handel a single field only you
        should use the model's method named like `on_change_FIELD_NAME`.
        These method will be processed automatically, based on th `on_change_prefix`
        attribute.
        """
        # for origin_name, origin_value, name, value in fields:
        #     pass  # handle logic here
        pass

    def post_change(self, fields: List[Tuple]):
        """
        This method allows to implement the logic based on the whole
        changed fields list. It works only with the field defined in the
        `post_change_fields` attribute. If you need to handel a single field only you
        should use the model's method named like `post_change_FIELD_NAME`.
        These method will be processed automatically, based on th `post_change_prefix`
        attribute.
        """
        # for origin_name, origin_value, name, value in fields:
        #     pass  # handle logic here
        pass

    def get_on_change_fields(self):
        # Combine manually added fields with the defined through method fields
        fields = self.get_change_methods(self.on_change_prefix)
        fields.extend(self.on_change_fields)
        return fields

    def get_post_change_fields(self):
        # Combine manually added fields with the defined through method fields
        fields = self.get_change_methods(self.post_change_prefix)
        fields.extend(self.post_change_fields)
        return fields

    def get_change_methods(self, prefix):
        def is_change_method(a):
            return a.startswith(prefix) and callable(getattr(self, a))
        l = len(prefix)
        return [attr[l:] for attr in dir(self) if is_change_method(attr)]

    def get_changed_fields(self, prefix) -> Set[Tuple]:
        if prefix == self.on_change_prefix:
            fields = self._on_change_fields
        elif prefix == self.post_change_prefix:
            fields = self._post_change_fields
        else:
            raise AttributeError("Wrong prefix")
        for name in set(fields):
            origin_name = self.get_origin_name(name)
            origin_value = self.get_origin_value(origin_name)
            value = getattr(self, name)
            if value != origin_value:
                yield origin_name, origin_value, name, value

    def get_on_change_counter(self):
        return self.get_origin_value(self.get_origin_name('counter'))

    @property
    def is_saved(self):
        return self.get_on_change_counter() > 0

    def _get_method(self, prefix, field_name: str):
        method_name = f'{prefix}{field_name}'
        try:
            return getattr(self, method_name)
        except AttributeError:
            # f'OnChangeModel has no method {method_name}'
            return

    def _increase_counter(self):
        counter = self.get_on_change_counter()
        setattr(self, self.get_origin_name('counter'), counter + 1)



# class MeViewSetMixin:
#     available_methods = []
#
#     def get_user(self, request):
#         return request.user
#
#     @action(methods=['get'])
#     def me(self, request, *args, **kwargs):
#         user = self.get_user(request)
#         if not user.is_authenticated:
#             return Response(status=status.404)
#         serializer = self.get_serializer(instance=user)
#         return Response(serializer.data)
#
#     @me.mark('patch')
#     def me_update(self, request, *args, **kwargs):
#         pass



import os
import json
# import base64

from cryptography.fernet import Fernet
# from cryptography.hazmat.backends import default_backend
# from cryptography.hazmat.primitives import hashes
# from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

from django.conf import settings
from rest_framework_simplejwt.tokens import RefreshToken


class TokenEncryption:
    @staticmethod
    def generate_salt(length: int):
        return os.urandom(length)

    @staticmethod
    def generate_key():
        return Fernet.generate_key()

    # @staticmethod
    # def k(data: dict, salt: bytes):
    #     kdf = PBKDF2HMAC(
    #         algorithm=hashes.SHA256(),
    #         length=32,
    #         salt=salt,
    #         iterations=100000,
    #         backend=default_backend(),
    #     )
    #     data = json.dumps(data).encode()
    #     return base64.urlsafe_b64encode(kdf.derive(data))

    @staticmethod
    def encrypt_data(data: dict, key: str = settings.JWT_PAYLOAD_ENCRYPTION_KEY):
        fernet = Fernet(key)
        data = json.dumps(data).encode()
        return fernet.encrypt(data)

    @staticmethod
    def decrypt_data(encrypted: bytes, key: str = settings.JWT_PAYLOAD_ENCRYPTION_KEY):
        fernet = Fernet(key)
        data = fernet.decrypt(encrypted)
        return json.loads(data)


class JWTUserModelMixin:
    def get_public_jwt_data(self) -> dict:
        """
        This method must return a payload dictionary. Fill this method with needed logic
        for building an open payload you want to put into the token.
        
        NOTICE! Be sure you do not use any sensitive data.
        JWT tokens could be unpacked by any person, so the information inside the token
        available for anyone who has this token.
        If you need to fill JWT token with a sensitive data, please use the 
        get_private_jwt_data method instead.
        
        :return: a payload dict
        """
        
        raise NotImplemented(
            "You should override this method with a specific payload builder."
        )
    
    def get_private_jwt_data(self) -> dict:
        """
        This method must return a payload dictionary. Fill this method with needed logic
        for building a sensitive data payload you want to put into the token.
        
        :return: a payload dict 
        """
        
        return {}

    def extend_token(self, token):
        """
        It takes a fresh token and extends it with an additional payload.
        There are two options: get_public_jwt_data and get_private_jwt_data. 
        :param token: a fresh valid RefreshToken
        :return: an extended RefreshToken
        """

        # Fill in the token with the public data
        public_data = self.get_public_jwt_data()
        for k, v in public_data.items():
            token[k] = v

        # Fill in the token with the private data
        private_data = self.get_private_jwt_data()
        token['private'] = TokenEncryption.encrypt_data(private_data).decode()
        return token

    def build_token(self):
        """
        It builds a RefreshToken instance and runs the extending process.
        :return: an extended RefreshToken instance
        """
        token = RefreshToken.for_user(self)
        token = self.extend_token(token)
        return token

    def get_tokens_pair(self) -> dict[str, str]:
        """
        Wrapper for the build_token method.
        :return: a dict with the access and  refresh tokens inside.
        """
        refresh = self.build_token()
        return {
            'refresh': str(refresh),
            'access': str(refresh.access_token),
        }



import signal
import os
from typing import Iterable

from django.core.management.commands.runserver import Command as RunServerOriginal
from django.utils.module_loading import import_string
from gears.settings import get_settings


class Command(RunServerOriginal):
    option_prefix: str = "run_gears_server_"

    def __init__(self):
        super().__init__()
        self.settings = get_settings()
        self.server_has_stopped = None

    def handle(self, *args, **options):
        # Set up signal handlers in the main thread
        signal.signal(signal.SIGINT, self.server_stop_handler)
        signal.signal(signal.SIGTERM, self.server_stop_handler)

        if options.get('use_reloader', True):
            # Check if our custom environment variable is set
            if os.environ.get("DJANGO_MAIN_PROCESS") != "True":  # main process
                os.environ["DJANGO_MAIN_PROCESS"] = "True"
                self.server_has_stopped = False
            else:  # chile process
                pass

        # Now call the super().handle which will start the server
        super().handle(*args, **options)

    def inner_run(self, *args, **options):
        self.server_start_handler()
        return super().inner_run(*args, **options)

    def server_start_handler(self, *args, **kwargs):
        tasks = self._get_tasks('up')
        self.stdout.write(
            self.style.SUCCESS(f'GEARS detected {len(tasks)} tasks to be '
                               f'run on the server`s starting up.')
        )
        self._run_tasks(tasks)

    def server_stop_handler(self, signum, frame):
        if self.server_has_stopped is False:  # never stopped before
            self.server_has_stopped = True

            tasks = self._get_tasks('down')
            self.stdout.write(
                self.style.SUCCESS(f'GEARS detected {len(tasks)} tasks to be '
                                   f'run on the server`s down.')
            )
            self._run_tasks(tasks)

        raise KeyboardInterrupt

    def _run_tasks(self, tasks: Iterable):
        for task, kw in tasks:
            self.stdout.write(self.style.SUCCESS(task))
            import_string(task)(**kw)

    def _get_tasks(self, t: str):
        # Example:
        # dict(
        #     run_gears_server_up_tasks=[
        #         # Dotted path to the function
        #         "path.to.task1",
        #     ],
        #     run_gears_server_down_tasks=[
        #         # Dotted path to the function with keyword arguments
        #         (
        #             "path.to.task2",
        #             {'a': "A", 'b': "B"},
        #         ),
        #     ],
        # )
        tasks = []
        for task in self.settings[f'{self.option_prefix}{t}_tasks']:
            kw = {}
            if isinstance(task, list) or isinstance(task, tuple):
                if len(task) != 2:
                    raise AttributeError(
                        "Run GEARS server tasks must be a set of dotted path of "
                        "functions or a set of pairs: dotted path to function "
                        "and list of kwargs."
                    )
                kw = task[1]
                task = task[0]
            tasks.append((task, kw,))
        return tasks



class SerializersMixin(object):
    """
    SerializersMixin ...
    """
    serializers = {}
    serializer_class = None
    default_name = 'default'
    _serializers = {
        default_name: serializer_class,
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.serializer_class and not self.serializers.get(self.default_name):
            raise ValueError(
                "Need to specify either 'serializer_class' or "
                "'serializers default' class"
            )
        self._serializers.update(self.serializers)

    def get_serializer_class(self, serializer_name=None):
        return self.serializers.get(serializer_name) \
               or self.serializers.get(self.action) \
               or self.serializers.get(self.default_name) \
               or super().get_serializer_class()

    def get_serializer(self, *args, **kwargs):
        serializer_class = self.get_serializer_class(
            serializer_name=kwargs.pop('serializer_name', None)
        )
        kwargs['context'] = self.get_serializer_context()
        return serializer_class(*args, **kwargs)



from rest_framework_simplejwt.serializers import TokenObtainPairSerializer


class JWTObtainPairSerializer(TokenObtainPairSerializer):
    """
    This is a helper serializer, which works with a JWTUserModelMixin.
    """

    @classmethod
    def get_token(cls, user):
        if not hasattr(user, 'build_token'):
            raise Exception(
                'JWTObtainPairSerializer must be used with a JWTUserModelMixin together'
            )
        return user.build_token()



from django.conf import settings

from .types import Error

DEFAULT_MAPPING = dict(
    required=Error(code='value_required'),
    unique=Error(code='not_unique_value'),
    method_not_allowed=Error(location='http'),
)

RESPONSE_ERROR_MAPPING = getattr(settings, 'RESPONSE_ERROR_MAPPING', None) or DEFAULT_MAPPING



import setuptools


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="drf-gears",
    version="0.9.15",
    author="Alexander Yudkin",
    author_email="san4ezy@gmail.com",
    description="Some gears collection for getting life a little bit better.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/san4ezy/drf-gears",
    packages=setuptools.find_packages(where='src'),
    package_dir={'': 'src'},
    include_package_data=True,
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    python_requires='>=3.7',
    install_requires=[
    ],
)



from django.contrib import admin
from django.urls import path


class SingletonModelAdmin(admin.ModelAdmin):
    def get_urls(self):
        urlpatterns = super().get_urls()
        try:
            model_name = self.model.__class__.__name__.lower()
            custom_urlpatterns = [
                path(r'',
                     self.admin_site.admin_view(self.change_view),
                     {'object_id': str(self.model.get().pk)},
                     name=f'{self.model._meta.app_label}_{model_name}_change',
                     ),
            ]
        except AttributeError as e:
            custom_urlpatterns = []
        return custom_urlpatterns + urlpatterns

    def has_add_permission(self, request):
        return False

    def has_delete_permission(self, request, obj=None):
        return False



from dataclasses import dataclass
from typing import Union


@dataclass
class Error:
    code: str = None
    location: str = None
    description: str = None
    detail: Union[dict[str, Union[str, int]], None] = None


@dataclass
class Response:
    success: bool
    status_code: int
    pagination: Union[dict, None]
    errors: Union[Error, list]
    data: Union[dict, list, None]
    service_data: Union[dict, list, None]



from copy import copy

from django.core.exceptions import PermissionDenied
from django.http import Http404, HttpResponseNotFound
from gears.renderers.mapping import RESPONSE_ERROR_MAPPING
from gears.renderers.types import Error
from rest_framework import exceptions
from rest_framework import status

from rest_framework.response import Response
from rest_framework.views import set_rollback


class ExceptionHandler(object):
    default_code: str = 'unknown_error'
    default_description: str = 'Unknown error'
    default_status_code: str = status.HTTP_500_INTERNAL_SERVER_ERROR
    root_field_name: str = 'errors'

    def __init__(self, exc, context):
        self.exc = exc
        self.context = context
        self.status_code = None
        self.errors: list = []
        self.data = {}
        self.headers = {}

    def _super(self):
        exc = self.exc
        headers = {}
        if isinstance(exc, Http404):
            exc = HttpResponseNotFound()
        elif isinstance(exc, PermissionDenied):
            exc = exceptions.PermissionDenied()

        try:
            self.status_code = exc.status_code
        except AttributeError:
            self.status_code = self.default_status_code

        if isinstance(exc, exceptions.APIException):
            if getattr(exc, 'auth_header', None):
                headers['WWW-Authenticate'] = exc.auth_header
            if getattr(exc, 'wait', None):
                headers['Retry-After'] = '%d' % exc.wait
            set_rollback()

            self.data = exc.detail
        else:
            self.data = str(exc)
        self.headers = headers
        self.exc = exc

    def process_list(self, detail: list, location):
        for v in detail:
            self.process(v, location=location)

    def process_dict(self, detail: dict, location):
        for k, v in detail.items():
            self.process(v, location=k)

    def process_error(self, detail, location):
        if status.is_server_error(self.status_code):
            raise self.exc
        self.errors.append(
            Error(
                code=getattr(detail, 'code', self.default_code),
                location=location,
                description=str(detail) or self.default_description,
                detail=None,
            )
        )

    def process(self, detail, location: str = None):
        if isinstance(detail, list):
            return self.process_list(detail, location)
        if isinstance(detail, dict):
            return self.process_dict(detail, location)
        return self.process_error(detail, location)

    def response(self) -> list[dict]:
        def _build(e: Error):
            er = copy(RESPONSE_ERROR_MAPPING.get(e.code, e))
            for k in ('code', 'location', 'description', 'detail',):
                # set values from me or e
                setattr(er, k, getattr(er, k) or getattr(e, k))
            return er.__dict__
        return [_build(err) for err in self.errors]

    def handle(self):
        self._super()
        self.process(self.data)
        data = {
            self.root_field_name: self.response(),
        }

        return Response(
            data=data,
            headers=self.headers,
            status=self.status_code,
        )


def exception_handler(exc, context):
    return ExceptionHandler(exc, context).handle()
